/*
 * Gateway for Wireless temperature and humidity sensors. Works with WT450H sensors and WS2015H display unit.
 *
 * v0.1
 * Jaakko Ala-Paavola
 * 2011-02-27
 */

#include <SPI.h>
#include <Ethernet.h>
#include <Udp.h>
#include <Time.h>


#define DEBUG 0
#define ETHERNET 1


#define TIMEOUT 1000
#define BIT0_LENGTH 2000
#define BIT1_LENGTH 1000
#define VARIATION 500
#define DATA_LENGTH 5
#define SENSOR_COUNT 4
#define NETWORK_COUNT 2
#define MSGLENGTH 36

#define NTP_SYNC_INTERVAL 60*60
#define NTP_REPLY_DELAY 2

//#undefine FLOATING_POINT

#ifdef FLOATING_POINT
struct sensor {
  unsigned char humidity;
  float         temp;
  float         temp_min;
  float         temp_max;
//  float         average;
  time_t        timestamp;
};
#else
struct sensor {
  unsigned char humidity;
  signed char   temp_int;
  unsigned char temp_dec;
  signed char   min_int;
  unsigned char min_dec;
  signed char   max_int;
  unsigned char max_dec;
  time_t timestamp;
};
#endif

byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };
byte ip[] = { 192, 168, 1, 177 };
byte gw[] = { 192, 168, 1, 1 };

unsigned int localPort = 8888;  
byte timeServer[] = { 192, 43, 244, 18}; // time.nist.gov NTP server
  
//byte SNTP_server_IP[]    = { 192, 43, 244, 18}; // time.nist.gov

Server server(80);

char date[10];
unsigned long time;
unsigned char bitcount = 0;
unsigned char bytecount = 0;
unsigned char second_half = 0;
unsigned char data[DATA_LENGTH];
struct sensor measurement[NETWORK_COUNT][SENSOR_COUNT];
unsigned char minmaxday = 0;
//unsigned long data;

const int NTP_PACKET_SIZE= 48; // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets 

time_t startTime = 0;
time_t lastSync = 0;
unsigned char ntpSyncRequested = 0;
unsigned int ntpSyncOccurred = 0;

//time_t prevDisplay = 0; // when the digital clock was displayed
//const  long timeZoneOffset = 0L; // set this to the offset in seconds to your local time;

void serialPrintTime(time_t t);
void serialPrintDate(time_t t);
unsigned long receiveNTPtime();
unsigned long sendNTPpacket(byte *address);
unsigned long epoch( unsigned long ntp_time );
void generateWebPage(Client client);
void webPrintTime(Client client, time_t t);
void webPrintDate(Client client, time_t t);

/* This funtion is executed when state change occurs in external interrupt 0 */
void receive()
{
    unsigned char i;
    unsigned char bit;
    unsigned long current = micros();
    unsigned long diff = current-time;
    time=current;
    
    time_t t;
    
    if ( diff < BIT0_LENGTH + VARIATION && diff > BIT0_LENGTH - VARIATION )
     {
        bit = 0;
        second_half = 0;
     }
     else if ( diff < BIT1_LENGTH + VARIATION && diff > BIT1_LENGTH - VARIATION )
     {
       if (second_half)
       {
         bit = 1;
         second_half = 0;  
       }
       else
       {
         second_half = 1;
         return;
       }
     }
     else 
     {
      goto reset;
     }
     
     data[bitcount/8] = data[bitcount/8]<<1;
     data[bitcount/8] |= bit;
     bitcount++;
     
     if ( bitcount == 4 )
     {
       if ( data[0] != 0x0c )
         goto reset;
       bitcount = 8;
#if DEBUG
       Serial.print('#');
#endif
     }

     if ( bitcount >= MSGLENGTH )
     {
       unsigned char net, id;
       int t_int;
       unsigned char rh, t_dec;

#if DEBUG       
       for (i=0; i<DATA_LENGTH; i++) {
         Serial.print(data[i], HEX);
         Serial.print(' ');
       }
#endif
       
       Serial.print("Data: NET:");
       net = 0x07 & (data[1]>>4)-1;
       Serial.print(net,DEC);
       Serial.print(" ID:");
       id = 0x03 & (data[1]>>2);
       Serial.print(id, DEC);
       Serial.print(" RH:");
       rh = data[2];
       Serial.print(rh, DEC);
       Serial.print(" T:");
       t_int = data[3]-50;
       Serial.print(t_int, DEC);
       t_dec = data[4]>>1;
       Serial.print('.');
       Serial.print(t_dec,DEC);

         
        t = now();

       Serial.print(" DATE:");
       serialPrintDate(t);
       
       Serial.print(" TIME:");
       serialPrintTime(t);
       
       Serial.print("\r\n");
      
       /* Do not store all sensors */
       
       if ( net <= NETWORK_COUNT )
       {
#ifdef FLOATING_POINT
         float tmp = t_int + (float)t_dec/10;

         measurement[net][id].temp = tmp;
         measurement[net][id].humidity = rh;
         
         if ( !measurement[net][id].timestamp ||
               day(t) != day(measurement[net][id].timestamp) )
         {
            measurement[net][id].temp_max = tmp;
            measurement[net][id].temp_min = tmp;
         }
         else
         {
           if ( tmp > measurement[net][id].temp_max )
            {
              measurement[net][id].temp_max = tmp;
            }

           if ( tmp < measurement[net][id].temp_min )
            {
              measurement[net][id].temp_min = tmp;
            }
         }
#else
         measurement[net][id].temp_int = t_int;
         measurement[net][id].temp_dec = t_dec;
         measurement[net][id].humidity = rh;
         
         if ( !measurement[net][id].timestamp ||
              day(t) != minmaxday )
         {
            minmaxday = day(t);
            measurement[net][id].max_int = t_int;
            measurement[net][id].max_dec = t_dec;  
            measurement[net][id].min_int = t_int;
            measurement[net][id].min_dec = t_dec;          
         }

         if ( t_int > measurement[net][id].max_int ||
              ( t_int == measurement[net][id].max_int && t_dec > measurement[net][id].max_dec ) )
          {
            measurement[net][id].max_int = t_int;
            measurement[net][id].max_dec = t_dec;
          }

         if ( t_int < measurement[net][id].min_int ||
              ( t_int == measurement[net][id].min_int && t_dec < measurement[net][id].min_dec ) )
          {
            measurement[net][id].min_int = t_int;
            measurement[net][id].min_dec = t_dec;
          }
       
#endif      
         measurement[net][id].timestamp = t;
       }
       
       goto reset;  
     }
    return;

reset: 
  for (i=0; i<DATA_LENGTH; i++)
    data[i] = 0;

   bytecount = 0;
   bitcount = 0;
   second_half = 0;
   return;
}

void setup()
{
  unsigned char i,j;
  unsigned long t_epoch;
  
  Serial.begin(115200);
  Serial.println("Info: Starting Wireless Sensors Gateway");
  
  for (j=0; j<NETWORK_COUNT; j++)
  {
    for (i=0; i<SENSOR_COUNT; i++)
    {
#ifdef FLOATING_POINT
      measurement[j][i].temp = 0.0;
      measurement[j][i].temp_max = 0.0;
      measurement[j][i].temp_min = 0.0;  
      measurement[j][i].humidity = 0;
      measurement[j][i].timestamp = 0;
#else
      measurement[j][i].temp_int = 0;
      measurement[j][i].temp_dec = 0;
      measurement[j][i].max_int = 0;
      measurement[j][i].max_dec = 0;      
      measurement[j][i].min_int = 0;
      measurement[j][i].min_dec = 0;  
      measurement[j][i].humidity = 0;
      measurement[j][i].timestamp = 0;
#endif    
    }
  }

#if ETHERNET  
  Serial.println("Info: Starting network interface");
  Ethernet.begin(mac, ip, gw);
  Udp.begin(localPort);

  Serial.println("Info: Requesting NTP time...");

  sendNTPpacket(timeServer);
  
  Serial.println("Info: Starting web service");
  
  server.begin();

#endif
  
  Serial.println("Info: IO initialization");
  attachInterrupt(0,receive,CHANGE);
  
  time = micros();

#if ETHERNET 
  delay(NTP_REPLY_DELAY*1000);
 
   t_epoch = epoch( receiveNTPtime() );

  if ( t_epoch )
  {
      setTime(t_epoch);
      
      lastSync = startTime = t_epoch;
      
      Serial.print("Info: Received NTP Time (UTC): ");
      serialPrintDate(t_epoch);
      Serial.print(' ');
      serialPrintTime(t_epoch);
      Serial.println(' ');
      
      ntpSyncOccurred++;
  }
  else
  {
      Serial.println("Error: Failed to receive NTP time");
  }
#endif

  Serial.println("Info: ### Setup done. Ready for operation. ###");
}

#define BUFFER_SIZE 6
void loop()
{
#if ETHERNET
  unsigned char i,j, k;
  time_t t;
  char buf[BUFFER_SIZE];
  char c;
  
    // listen for incoming clients
  Client client = server.available();
  if (client) {
    // an http request ends with a blank line
    boolean currentLineIsBlank = true;
    k = 0;
    while (client.connected()) {
      if (client.available()) {
        char c = client.read();
        if ( k < BUFFER_SIZE )
        {
          buf[k] = c;
          k++;
        }
        // if you've gotten to the end of the line (received a newline
        // character) and the line is blank, the http request has ended,
        // so you can send a reply
        if (c == '\n' && currentLineIsBlank) {
          // send a standard http response header

           if ( buf[0] == 'G' && buf[1] == 'E' && buf[2] == 'T' && buf[3] == ' ' && buf[4] == '/' )
           {
              buf[BUFFER_SIZE-1] = 0;
              if ( buf[5] == ' ' )
              {
                  Serial.println("Info: Requested main web page.");
                  generateWebPage(client);
              }
              else
              {
                Serial.print("Info: Requested web page: ");
                Serial.println(buf+4);
              }
           }
          break;
        }
        if (c == '\n') {
          // you're starting a new line
          currentLineIsBlank = true;
        }
        else if (c != '\r') {
          // you've gotten a character on the current line
          currentLineIsBlank = false;
        }
      }
    }
    // give the web browser time to receive the data
    delay(1);
    // close the connection:
    client.stop();
  }

  t = now();
    
  if ( !ntpSyncRequested && t-lastSync > NTP_SYNC_INTERVAL )
  {
    Serial.println("Info: Requesting NTP sync..");
    sendNTPpacket(timeServer);
    ntpSyncRequested = 1;
  }
  if ( ntpSyncRequested && t-lastSync > NTP_SYNC_INTERVAL + NTP_REPLY_DELAY )
  {
    unsigned long t_epoch = epoch( receiveNTPtime() );
    if ( t_epoch )
    {
      setTime(t_epoch);
      Serial.print("Info: NTP sync, diff ");
      Serial.print(t-t_epoch);
      Serial.print(" secs, new time: ");
      serialPrintTime(t_epoch);
      Serial.println(' ');
      ntpSyncRequested = 0;
      lastSync = t_epoch;
      ntpSyncOccurred++;
    }  
    else
    {
      Serial.print("Error: Failed to sync NTP time");
    }
  }
#endif // ETHERNET  
}
 
/*--- Time ---*/

void serialPrintTime(time_t t)
{
  Serial.print(hour(t),DEC);
  Serial.print(':');
  if ( minute(t) < 10 ) Serial.print('0');
  Serial.print(minute(t),DEC);
  Serial.print(':');
  if ( second(t) < 10 ) Serial.print('0');
  Serial.print(second(t),DEC);
}

void serialPrintDate(time_t t)
{
  Serial.print(year(t),DEC);
  Serial.print('-');
  if ( month(t) < 10 ) Serial.print('0');
  Serial.print(month(t),DEC);
  Serial.print('-');
  if ( day(t) < 10 ) Serial.print('0'); 
  Serial.print(day(t),DEC);
}




